
(*s Une position est reprsente par un entier $(a,b)$, dont
les bits codent la prsence ou non d'une bille.  Le codage est le
suivant :
    \begin{displaymath}
      \begin{array}{c|ccccccc}
        & 0      & 1      & 2 & 3      & 4 & 5 & 6 \\\hline
      0 & a_0    & a_1    &   & \dots  &   &   & a_6 \\
      1 & a_7    & a_8    &   & \dots  &   &   & a_{13} \\
      2 & a_{14} & a_{15} &   & \dots  &   &   & a_{20} \\
      3 & a_{21} & a_{22} &   & \dots  &   &   & a_{27} \\
      4 & b_0    & b_1    &   & \dots  &   &   & b_6 \\
      5 & b_7    & b_8    &   & \dots  &   &   & b_{13} \\
      6 & a_{14} & b_{15} &   & \dots  &   &   & b_{20}
    \end{array}
    \end{displaymath}
 *)

type position = int * int

(*s On hash les positions de manire efficace, avec le module
spcialis prcdemment dfini. *)

module HashPos = Hashtbl

(*s Pour simplifier la manipulation de ces bits par la suite, on cre
une matrice $case(i,j)$ qui retourne le couple d'entier reprsentant
exactement la case $(i,j)$, pour $i,j\in\{0,\dots,6\}$. *)

let case = 
  let m = Array.make_matrix 7 7 (0,0) in
  for i = 0 to 6 do
    for j = 0 to 6 do
      m.(i).(j) <- 
        if j < 4 then (1 lsl (i + 7*j),0) else (0,1 lsl (i + 7*(j-4)))
    done
  done;
  m

(*s Les positions de dpart et d'arrive sont dfinies facilement. *)

let arrivee = case.(3).(3)

let depart = (0b1110111111111100111000011100, 0b001110000111001111111)

(*s Affichage d'une position. *)

let affiche (a,b) =
  for j = 0 to 6 do
    for i = 0 to 6 do
      let (a',b') = case.(i).(j) in
      if (a land a' == a') && (b land b' == b') then
	print_string "X " else print_string ". "
    done;
    print_newline()
  done

(*s Test d'arrt sur une position : il ne reste qu'une bille. (Utilis
    pour rechercher la premire solution plutt que la << vraie >>
    solution du solitaire.) Attention : [un_bit] est correct parce que
    l'on sait que [n > 0]. *)

let rec un_bit n = n == 1 || (n land 1 == 0 && un_bit (n lsr 1))

let test_arret (a,b) = ((un_bit a) && b == 0) || ((un_bit b) && a == 0)

(*s Un coup est reprsent par deux << positions >>, la premire
donnant les deux cases devant contenir une bille, la deuxime la case
devant tre vide. *)

type coup = position * position

(*s Ds lors, tester qu'un coup est possible dans une position donne
revient  tester les prsence des deux billes et de la case vide par
simples oprations de filtrage sur les bits.
*)

let est_possible (p1,p2) ((b1,b2),(v1,v2)) =
    (p1 land b1 == b1)
  && (p2 land b2 == b2)
  && ((lnot p1) land v1 == v1)
  && ((lnot p2) land v2 == v2)

(*s Jouer un coup revient  appliquer le ngatif de ces deux filtres
($\bullet\bullet\circ \rightarrow \circ\circ\bullet$). *)

let joue (p1,p2) ((b1,b2),(v1,v2)) =
  ((p1 lxor b1) lor v1, (p2 lxor b2) lor v2)

(*s La liste de tous les coups. *)

let tous_les_coups =
  let l = ref [] in
  (* ajoute un coup et son inverse *)
  let ajoute (i1,j1) (i2,j2) (i3,j3) =
    let (a1,b1) as c1 = case.(i1).(j1) in
    let (a2,b2) = case.(i2).(j2) in
    let (a3,b3) as c3 = case.(i3).(j3) in
    l := ((a1 lor a2, b1 lor b2), c3) ::
         ((a2 lor a3, b2 lor b3), c1) :: !l
  in
  for j = 0 to 6 do
    ajoute (2,j) (3,j) (4,j);
    ajoute (j,2) (j,3) (j,4)
  done;
  for i = 0 to 4 do
    ajoute (2,i) (2,i+1) (2,i+2);
    ajoute (3,i) (3,i+1) (3,i+2);
    ajoute (4,i) (4,i+1) (4,i+2);
    ajoute (i,2) (i+1,2) (i+2,2);
    ajoute (i,3) (i+1,3) (i+2,3);
    ajoute (i,4) (i+1,4) (i+2,4)
  done;
  !l

(*s Affichage d'une solution. *)

let affiche_coup ((b1,b2),cv) =
  let iv = ref 0 and jv = ref 0 in
  for i = 0 to 6 do
    for j = 0 to 6 do
      if case.(i).(j) = cv then begin iv := i; jv := j end
    done
  done;
  let dir = ref "" in
  for i = 0 to 6 do
    for j = 0 to 6 do
      let (c1,c2) = case.(i).(j) in 
      if (b1 land c1 <> 0) || (b2 land c2 <> 0) then
	if i < !iv then dir := "gauche"
	else if i > !iv then dir := "droite"
	else if j > !jv then dir := "bas"
	else dir := "haut"
    done
  done;
  Printf.printf "(%d, %d) %s\n" !iv !jv !dir
      
let affiche_solution l =
  List.iter affiche_coup (List.rev l)
